@@ -24,18 +24,22 @@
 #
 ###############################################################################
 
-from __future__ import absolute_import
+import copy
+import math
+from typing import Optional
 
 import txaio
 
 from twisted.internet.protocol import Factory
 from twisted.protocols.basic import Int32StringReceiver
 from twisted.internet.error import ConnectionDone
+from twisted.internet.defer import CancelledError
 
-from autobahn.util import public
-from autobahn.twisted.util import peer2str, transport_channel_id
-from autobahn.util import _LazyHexFormatter
-from autobahn.wamp.exception import ProtocolError, SerializationError, TransportLost
+from autobahn.util import public, _LazyHexFormatter
+from autobahn.twisted.util import create_transport_details, transport_channel_id
+from autobahn.wamp.types import TransportDetails
+from autobahn.wamp.exception import ProtocolError, SerializationError, TransportLost, InvalidUriError
+from autobahn.exception import PayloadExceededError
 
 __all__ = (
     'WampRawSocketServerProtocol',
@@ -51,18 +55,37 @@ class WampRawSocketProtocol(Int32StringReceiver):
     """
     log = txaio.make_logger()
 
-    def connectionMade(self):
-        self.log.debug("WampRawSocketProtocol: connection made")
+    peer: Optional[str] = None
+    is_server: Optional[bool] = None
 
-        # the peer we are connected to
-        #
-        try:
-            peer = self.transport.getPeer()
-        except AttributeError:
-            # ProcessProtocols lack getPeer()
-            self.peer = "?"
-        else:
-            self.peer = peer2str(peer)
+    def __init__(self):
+        # set the RawSocket maximum message size by default
+        self._max_message_size = 2**24
+        self._transport_details = None
+
+    @property
+    def transport_details(self) -> Optional[TransportDetails]:
+        """
+        Implements :func:`autobahn.wamp.interfaces.ITransport.transport_details`
+        """
+        return self._transport_details
+
+    def lengthLimitExceeded(self, length):
+        # override hook in Int32StringReceiver base class that is fired when a message is (to be) received
+        # that is larger than what we agreed to handle (by negotiation in the RawSocket opening handshake)
+        emsg = 'RawSocket connection: length of received message exceeded (message was {} bytes, but current maximum is {} bytes)'.format(length, self.MAX_LENGTH)
+        raise PayloadExceededError(emsg)
+
+    def connectionMade(self):
+        # Twisted networking framework entry point, called by Twisted
+        # when the connection is established (either a client or a server)
+
+        # determine preliminary transport details (what is know at this point)
+        self._transport_details = create_transport_details(self.transport, self.is_server)
+        self._transport_details.channel_framing = TransportDetails.CHANNEL_FRAMING_RAWSOCKET
+
+        # backward compatibility
+        self.peer = self._transport_details.peer
 
         # a Future/Deferred that fires when we hit STATE_CLOSED
         self.is_closed = txaio.create_future()
@@ -85,23 +108,37 @@ class WampRawSocketProtocol(Int32StringReceiver):
         #
         self._handshake_bytes = b''
 
-        # Client requested maximum length of serialized messages.
+        # Peer requested to _receive_ this maximum length of serialized messages - hence we must not send larger msgs!
         #
         self._max_len_send = None
 
     def _on_handshake_complete(self):
+        # RawSocket connection established. Now let the user WAMP session factory
+        # create a new WAMP session and fire off session open callback.
         try:
+            if self._transport_details.is_secure:
+                # now that the TLS opening handshake is complete, the actual TLS channel ID
+                # will be available. make sure to set it!
+                channel_id = {
+                    'tls-unique': transport_channel_id(self.transport, self._transport_details.is_server, 'tls-unique'),
+                }
+                self._transport_details.channel_id = channel_id
+
             self._session = self.factory._factory()
-            self._session.onOpen(self)
+            self.log.debug('{klass}._on_handshake_complete(): calling {method}', session=self._session,
+                           klass=self.__class__.__name__, method=self._session.onOpen)
+            res = self._session.onOpen(self)
         except Exception as e:
             # Exceptions raised in onOpen are fatal ..
-            self.log.warn("WampRawSocketProtocol: ApplicationSession constructor / onOpen raised ({err})", err=e)
+            self.log.warn("{klass}._on_handshake_complete(): ApplicationSession constructor / onOpen raised ({err})",
+                          klass=self.__class__.__name__, err=e)
             self.abort()
         else:
-            self.log.debug("ApplicationSession started.")
+            self.log.debug('{klass}._on_handshake_complete(): {session} started (res={res}).', klass=self.__class__.__name__,
+                           session=self._session, res=res)
 
     def connectionLost(self, reason):
-        self.log.debug("WampRawSocketProtocol: connection lost: reason = '{reason}'", reason=reason)
+        self.log.debug('{klass}.connectionLost(reason="{reason}"', klass=self.__class__.__name__, reason=reason)
         txaio.resolve(self.is_closed, self)
         try:
             wasClean = isinstance(reason.value, ConnectionDone)
@@ -109,22 +146,53 @@ class WampRawSocketProtocol(Int32StringReceiver):
                 self._session.onClose(wasClean)
         except Exception as e:
             # silently ignore exceptions raised here ..
-            self.log.warn("WampRawSocketProtocol: ApplicationSession.onClose raised ({err})", err=e)
+            self.log.warn('{klass}.connectionLost(): ApplicationSession.onClose raised "{err}"',
+                          klass=self.__class__.__name__, err=e)
         self._session = None
 
     def stringReceived(self, payload):
-        self.log.trace("WampRawSocketProtocol: RX octets: {octets}", octets=_LazyHexFormatter(payload))
+        self.log.trace('{klass}.stringReceived(): RX {octets} octets',
+                       klass=self.__class__.__name__, octets=_LazyHexFormatter(payload))
         try:
             for msg in self._serializer.unserialize(payload):
-                self.log.trace("WampRawSocketProtocol: RX WAMP message: {msg}", msg=msg)
+                self.log.trace("{klass}.stringReceived: RX WAMP message: {msg}",
+                               klass=self.__class__.__name__, msg=msg)
                 self._session.onMessage(msg)
 
+        except CancelledError as e:
+            self.log.debug("{klass}.stringReceived: WAMP CancelledError - connection will continue!\n{err}",
+                           klass=self.__class__.__name__,
+                           err=e)
+
+        except InvalidUriError as e:
+            self.log.warn("{klass}.stringReceived: WAMP InvalidUriError - aborting connection!\n{err}",
+                          klass=self.__class__.__name__,
+                          err=e)
+            self.abort()
+
         except ProtocolError as e:
-            self.log.warn("WampRawSocketProtocol: WAMP Protocol Error ({err}) - aborting connection", err=e)
+            self.log.warn("{klass}.stringReceived: WAMP ProtocolError - aborting connection!\n{err}",
+                          klass=self.__class__.__name__,
+                          err=e)
+            self.abort()
+
+        except PayloadExceededError as e:
+            self.log.warn("{klass}.stringReceived: WAMP PayloadExceededError - aborting connection!\n{err}",
+                          klass=self.__class__.__name__,
+                          err=e)
+            self.abort()
+
+        except SerializationError as e:
+            self.log.warn("{klass}.stringReceived: WAMP SerializationError - aborting connection!\n{err}",
+                          klass=self.__class__.__name__,
+                          err=e)
             self.abort()
 
         except Exception as e:
-            self.log.warn("WampRawSocketProtocol: WAMP Internal Error ({err}) - aborting connection", err=e)
+            self.log.failure()
+            self.log.warn("{klass}.stringReceived: WAMP Exception - aborting connection!\n{err}",
+                          klass=self.__class__.__name__,
+                          err=e)
             self.abort()
 
     def send(self, msg):
@@ -132,15 +200,24 @@ class WampRawSocketProtocol(Int32StringReceiver):
         Implements :func:`autobahn.wamp.interfaces.ITransport.send`
         """
         if self.isOpen():
-            self.log.trace("WampRawSocketProtocol (serializer={serializer}): TX WAMP message: {msg}", msg=msg, serializer=self._serializer)
+            self.log.trace('{klass}.send() (serializer={serializer}): TX WAMP message: "{msg}"',
+                           klass=self.__class__.__name__, msg=msg, serializer=self._serializer)
             try:
                 payload, _ = self._serializer.serialize(msg)
-            except Exception as e:
+            except SerializationError as e:
                 # all exceptions raised from above should be serialization errors ..
                 raise SerializationError("WampRawSocketProtocol: unable to serialize WAMP application payload ({0})".format(e))
             else:
-                self.sendString(payload)
-                self.log.trace("WampRawSocketProtocol: TX octets: {octets}", octets=_LazyHexFormatter(payload))
+                payload_len = len(payload)
+                if 0 < self._max_len_send < payload_len:
+                    emsg = 'tried to send RawSocket message with size {} exceeding payload limit of {} octets'.format(
+                        payload_len, self._max_len_send)
+                    self.log.warn(emsg)
+                    raise PayloadExceededError(emsg)
+                else:
+                    self.sendString(payload)
+                    self.log.trace('{klass}.send(): TX {octets} octets',
+                                   klass=self.__class__.__name__, octets=_LazyHexFormatter(payload))
         else:
             raise TransportLost()
 
@@ -194,23 +271,28 @@ class WampRawSocketServerProtocol(WampRawSocketProtocol):
             if len(self._handshake_bytes) == 4:
 
                 self.log.debug(
-                    "WampRawSocketProtocol: opening handshake received - {octets}",
+                    "WampRawSocketServerProtocol: opening handshake received - 0x{octets}",
                     octets=_LazyHexFormatter(self._handshake_bytes),
                 )
 
-                if ord(self._handshake_bytes[0:1]) != 0x7f:
+                # first octet must be magic octet 0x7f
+                #
+                _magic = ord(self._handshake_bytes[0:1])
+                if _magic != 127:
                     self.log.warn(
-                        "WampRawSocketProtocol: invalid magic byte (octet 1) in"
-                        " opening handshake: was 0x{magic}, but expected 0x7f",
-                        magic=_LazyHexFormatter(self._handshake_bytes[0]),
+                        "WampRawSocketServerProtocol: invalid magic byte (octet 1) in"
+                        " opening handshake: was {magic}, but expected 127",
+                        magic=_magic,
                     )
                     self.abort()
+                else:
+                    self.log.debug('WampRawSocketServerProtocol: correct magic byte received')
 
                 # peer requests us to send messages of maximum length 2**max_len_exp
                 #
                 self._max_len_send = 2 ** (9 + (ord(self._handshake_bytes[1:2]) >> 4))
                 self.log.debug(
-                    "WampRawSocketProtocol: client requests us to send out most {max_bytes} bytes per message",
+                    "WampRawSocketServerProtocol: client requests us to send out most {max_bytes} bytes per message",
                     max_bytes=self._max_len_send,
                 )
 
@@ -218,22 +300,26 @@ class WampRawSocketServerProtocol(WampRawSocketProtocol):
                 #
                 ser_id = ord(self._handshake_bytes[1:2]) & 0x0F
                 if ser_id in self.factory._serializers:
-                    self._serializer = self.factory._serializers[ser_id]
+                    self._serializer = copy.copy(self.factory._serializers[ser_id])
                     self.log.debug(
-                        "WampRawSocketProtocol: client wants to use serializer '{serializer}'",
+                        "WampRawSocketServerProtocol: client wants to use serializer '{serializer}'",
                         serializer=ser_id,
                     )
                 else:
                     self.log.warn(
-                        "WampRawSocketProtocol: opening handshake - no suitable serializer found (client requested {serializer}, and we have {serializers}",
+                        "WampRawSocketServerProtocol: opening handshake - no suitable serializer found (client requested {serializer}, and we have {serializers}",
                         serializer=ser_id,
                         serializers=self.factory._serializers.keys(),
                     )
                     self.abort()
 
-                # we request the peer to send message of maximum length 2**reply_max_len_exp
+                # we request the client to send message of maximum length 2**reply_max_len_exp
                 #
-                reply_max_len_exp = 24
+                reply_max_len_exp = int(math.ceil(math.log(self._max_message_size, 2)))
+
+                # this is an instance attribute on the Twisted base class for maximum size
+                # of _received_ messages
+                self.MAX_LENGTH = 2**reply_max_len_exp
 
                 # send out handshake reply
                 #
@@ -248,7 +334,7 @@ class WampRawSocketServerProtocol(WampRawSocketProtocol):
                 self._on_handshake_complete()
 
                 self.log.debug(
-                    "WampRawSocketProtocol: opening handshake completed: {serializer}",
+                    "WampRawSocketServerProtocol: opening handshake completed: {serializer}",
                     serializer=self._serializer,
                 )
 
@@ -258,12 +344,6 @@ class WampRawSocketServerProtocol(WampRawSocketProtocol):
             if data:
                 self.dataReceived(data)
 
-    def get_channel_id(self, channel_id_type=u'tls-unique'):
-        """
-        Implements :func:`autobahn.wamp.interfaces.ITransport.get_channel_id`
-        """
-        return transport_channel_id(self.transport, is_server=True, channel_id_type=channel_id_type)
-
 
 @public
 class WampRawSocketClientProtocol(WampRawSocketProtocol):
@@ -277,13 +357,16 @@ class WampRawSocketClientProtocol(WampRawSocketProtocol):
 
     def connectionMade(self):
         WampRawSocketProtocol.connectionMade(self)
-        self._serializer = self.factory._serializer
+        self._serializer = copy.copy(self.factory._serializer)
 
-        # we request the peer to send message of maximum length 2**reply_max_len_exp
-        #
-        request_max_len_exp = 24
+        # we request the peer to send messages of maximum length 2**reply_max_len_exp
+        request_max_len_exp = int(math.ceil(math.log(self._max_message_size, 2)))
 
-        # send out handshake reply
+        # this is an instance attribute on the Twisted base class for maximum size
+        # of _received_ messages
+        self.MAX_LENGTH = 2**request_max_len_exp
+
+        # send out handshake request
         #
         request_octet2 = bytes(bytearray([
             ((request_max_len_exp - 9) << 4) | self._serializer.RAWSOCKET_SERIALIZER_ID]))
@@ -302,22 +385,22 @@ class WampRawSocketClientProtocol(WampRawSocketProtocol):
             if len(self._handshake_bytes) == 4:
 
                 self.log.debug(
-                    "WampRawSocketProtocol: opening handshake received - {handshake}",
+                    "WampRawSocketClientProtocol: opening handshake received - {handshake}",
                     handshake=_LazyHexFormatter(self._handshake_bytes),
                 )
 
                 if ord(self._handshake_bytes[0:1]) != 0x7f:
                     self.log.debug(
-                        "WampRawSocketProtocol: invalid magic byte (octet 1) in opening handshake: was 0x{magic}, but expected 0x7f",
+                        "WampRawSocketClientProtocol: invalid magic byte (octet 1) in opening handshake: was 0x{magic}, but expected 0x7f",
                         magic=_LazyHexFormatter(self._handshake_bytes[0]),
                     )
                     self.abort()
 
-                # peer requests us to send messages of maximum length 2**max_len_exp
+                # peer requests us to _send_ messages of maximum length 2**max_len_exp
                 #
                 self._max_len_send = 2 ** (9 + (ord(self._handshake_bytes[1:2]) >> 4))
                 self.log.debug(
-                    "WampRawSocketProtocol: server requests us to send out most {max} bytes per message",
+                    "WampRawSocketClientProtocol: server requests us to send out most {max} bytes per message",
                     max=self._max_len_send,
                 )
 
@@ -326,7 +409,7 @@ class WampRawSocketClientProtocol(WampRawSocketProtocol):
                 ser_id = ord(self._handshake_bytes[1:2]) & 0x0F
                 if ser_id != self._serializer.RAWSOCKET_SERIALIZER_ID:
                     self.log.error(
-                        "WampRawSocketProtocol: opening handshake - no suitable serializer found (server replied {serializer}, and we requested {serializers})",
+                        "WampRawSocketClientProtocol: opening handshake - no suitable serializer found (server replied {serializer}, and we requested {serializers})",
                         serializer=ser_id,
                         serializers=self._serializer.RAWSOCKET_SERIALIZER_ID,
                     )
@@ -337,7 +420,7 @@ class WampRawSocketClientProtocol(WampRawSocketProtocol):
                 self._on_handshake_complete()
 
                 self.log.debug(
-                    "WampRawSocketProtocol: opening handshake completed (using serializer {serializer})",
+                    "WampRawSocketClientProtocol: opening handshake completed (using serializer {serializer})",
                     serializer=self._serializer,
                 )
 
@@ -347,17 +430,47 @@ class WampRawSocketClientProtocol(WampRawSocketProtocol):
             if data:
                 self.dataReceived(data)
 
-    def get_channel_id(self, channel_id_type=u'tls-unique'):
-        """
-        Implements :func:`autobahn.wamp.interfaces.ITransport.get_channel_id`
-        """
-        return transport_channel_id(self.transport, is_server=False, channel_id_type=channel_id_type)
-
 
 class WampRawSocketFactory(Factory):
     """
     Base class for Twisted-based WAMP-over-RawSocket factories.
     """
+    log = txaio.make_logger()
+
+    def __init__(self, factory):
+        """
+
+        :param factory: A callable that produces instances that implement
+            :class:`autobahn.wamp.interfaces.ITransportHandler`
+        :type factory: callable
+        """
+        if callable(factory):
+            self._factory = factory
+        else:
+            self._factory = lambda: factory
+
+        # RawSocket max payload size is 16M (https://wamp-proto.org/_static/gen/wamp_latest_ietf.html#handshake)
+        self._max_message_size = 2**24
+
+    def resetProtocolOptions(self):
+        self._max_message_size = 2**24
+
+    def setProtocolOptions(self, maxMessagePayloadSize=None):
+        self.log.debug('{klass}.setProtocolOptions(maxMessagePayloadSize={maxMessagePayloadSize})',
+                       klass=self.__class__.__name__, maxMessagePayloadSize=maxMessagePayloadSize)
+        assert maxMessagePayloadSize is None or (type(maxMessagePayloadSize) == int and maxMessagePayloadSize >= 512 and maxMessagePayloadSize <= 2**24)
+        if maxMessagePayloadSize is not None and maxMessagePayloadSize != self._max_message_size:
+            self._max_message_size = maxMessagePayloadSize
+
+    def buildProtocol(self, addr):
+        self.log.debug('{klass}.buildProtocol(addr={addr})', klass=self.__class__.__name__, addr=addr)
+        p = self.protocol()
+        p.factory = self
+        p.MAX_LENGTH = self._max_message_size
+        p._max_message_size = self._max_message_size
+        self.log.debug('{klass}.buildProtocol() -> proto={proto}, max_message_size={max_message_size}, MAX_LENGTH={MAX_LENGTH}',
+                       klass=self.__class__.__name__, proto=p, max_message_size=p._max_message_size, MAX_LENGTH=p.MAX_LENGTH)
+        return p
 
 
 @public
@@ -380,10 +493,7 @@ class WampRawSocketServerFactory(WampRawSocketFactory):
         :type serializers: list of objects implementing
             :class:`autobahn.wamp.interfaces.ISerializer`
         """
-        if callable(factory):
-            self._factory = factory
-        else:
-            self._factory = lambda: factory
+        WampRawSocketFactory.__init__(self, factory)
 
         if serializers is None:
             serializers = []
@@ -448,10 +558,10 @@ class WampRawSocketClientFactory(WampRawSocketFactory):
            this list: CBOR, MessagePack, UBJSON, JSON).
         :type serializer: object implementing :class:`autobahn.wamp.interfaces.ISerializer`
         """
-        if callable(factory):
-            self._factory = factory
-        else:
-            self._factory = lambda: factory
+        WampRawSocketFactory.__init__(self, factory)
+
+        # Reduce the factory logs noise
+        self.noisy = False
 
         if serializer is None:
 
